#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <windows.h>
#include <vector>

#include <xaudio2.h>
#include <xaudio2fx.h>

#pragma comment(lib, "xaudio2.lib")
#pragma comment(lib, "winmm.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "ole32.lib")

int main() {
	enum {
		Bits         = (sizeof(float) * 8),
		Channel      = (2),
		SampleRate   = (44100),
		Align        = ((Channel * Bits) / 8),
		BytePerSec   = (SampleRate * Align),
		MaxVoice     = 32,
	};
	WAVEFORMATEX wfx = { WAVE_FORMAT_IEEE_FLOAT, Channel, SampleRate, BytePerSec, Align, Bits, 0 };

	IXAudio2 *pXAudio2 = nullptr;
	IXAudio2MasteringVoice *pMasterVoice = nullptr;
	IXAudio2SourceVoice *vSourceVoices[MaxVoice] = {};
	std::vector<float> vWaveData;

	CoInitializeEx( NULL, COINIT_MULTITHREADED ) ;
	XAudio2Create(&pXAudio2, 0, XAUDIO2_DEFAULT_PROCESSOR);
	pXAudio2->CreateMasteringVoice(&pMasterVoice);
	for(int i = 0; i < MaxVoice; i++) {
		IXAudio2SourceVoice* pSourceVoice = nullptr;
		pXAudio2->CreateSourceVoice(&pSourceVoice, (WAVEFORMATEX *)&wfx, 0, XAUDIO2_MAX_FREQ_RATIO);
		vSourceVoices[i] = pSourceVoice;
	}
	
	auto Reset = [&]() {
		for(auto & ref : vSourceVoices) {
			ref->Stop(0);
			ref->FlushSourceBuffers();
		}
	};

	auto Play = [&](int pitch) {
		static int index = 0;
		auto ref = vSourceVoices[index++ % MaxVoice];
		XAUDIO2_BUFFER buffer = {};
		buffer.pAudioData = (const BYTE *)vWaveData.data();
		buffer.AudioBytes = static_cast<UINT32>(vWaveData.size() * sizeof(float));

		ref->Stop(0);
		ref->FlushSourceBuffers();
		ref->SubmitSourceBuffer(&buffer);
		ref->SetFrequencyRatio(pow(2.0f, float(pitch) / 12.0f) / 11.0f);
		ref->Start(0);
	};

	auto UpdateWaveData = [&](uint16_t seed = 0x7FFF) {
		auto frand = []() {
			return (float(rand()) / float(0x7FFF)) * 2.0f - 1.0f;
		};
		auto triangle = [](float x, float d = 0.5f) {
			x += d;
			return (x - floor(x)) * 2.0f - 1.0f;
		};
		auto square = [&](float x, float d = 0.5f) {
			return triangle(x) > d ? 1.0f : -1.0f;
		};

		Reset();
		vWaveData.clear();

		int detune = ((seed & 0x0300) >>  8) + 1;
		int oct    = ((seed & 0x0C00) >> 10) + 1;
		int noize  = ((seed & 0x1000) >> 12);
		printf("oct = %d, detune = %d, noize = %d\n", oct, detune, noize);
		float Volume = 1.0f;
		
		size_t SampleNumber = SampleRate;
		
		float VolumeDecay = 1.0f / float(SampleNumber);
		float phase = 0.0;
		float Lv = 0.5f + frand() * 0.2f;
		float Rv = 0.5f + frand() * 0.2f;
		for(int i = 0 ; i < SampleNumber; i++) {
			float base_freq = 110.0f;
			if(noize) base_freq += frand() * 5.0f;
			float phase_delta = base_freq / float(SampleRate);
			phase += phase_delta;
			float data = sin(phase * 2.0f * 3.141592f);

			float temp_phase = phase;
			for(int i = 0; i < oct; i++) {
				float detfact = 1.0;
				for(int di = 0; di < detune; di++) {
					if(seed & 0x0001) data += square(temp_phase * detfact, 0.25f);
					if(seed & 0x0002) data += triangle(temp_phase * detfact, 0.75f);
					if(seed & 0x0004) data += sin(phase * 2.0f * 3.141592f * detfact);
					detfact *= 1.01f;
				}
				temp_phase *= 2.005f;
			}
			data /= float(oct);

			data *= Volume;
			vWaveData.push_back(data * Lv); //L
			vWaveData.push_back(data * Rv); //R
			Volume -= VolumeDecay;
		}
	};

	auto base = 0x20;
	UpdateWaveData();
	printf("SOUND ON : A S D F Keys\nCHANGE SOUND TONE PRESET : SPACE KEY\nQUIT : ESCAPE KEY\n");

	while((GetAsyncKeyState(VK_ESCAPE) & 0x8000) == 0) {
		if( GetAsyncKeyState('A') & 0x0001) Play(base + 0 - 12);
		if( GetAsyncKeyState('S') & 0x0001) Play(base + 4);
		if( GetAsyncKeyState('D') & 0x0001) Play(base + 7);
		if( GetAsyncKeyState('F') & 0x0001) Play(base + 11);
		if( GetAsyncKeyState(VK_UP) & 0x0001) base += 1;
		if( GetAsyncKeyState(VK_DOWN) & 0x0001) base -= 1;
		if( GetAsyncKeyState(VK_SPACE) & 0x0001) {
			auto presetNum = rand();
			printf("Set Preset Num = %08X\n", presetNum);
			UpdateWaveData(presetNum);
		}
		Sleep(16);
	}
	pMasterVoice->DestroyVoice();
	pXAudio2->Release();
	CoUninitialize();
	return 0;
}
